/*
 * Title:        sag_internal/convert_to.hpp
 * Description:  Implementation of convert_to and get
 * $Copyright (c) 2015-2018, 2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: convert_to.hpp 393681 2021-08-05 13:53:08Z matj $
 */


#ifndef _SAG_INTERNAL_CONVERT_TO_HPP_
#define _SAG_INTERNAL_CONVERT_TO_HPP_

// already in namespace com.softwareag.connectivity


namespace get_details {

	/** \internal
	 * Tag of types which can be used with get.
	 */

	template<typename T>
	using is_gettable = ampl::is_one_of<T, 
			int64_t, double, bool, // primatives
			std::string, const char *, // string
			map_t, list_t,	decimal_t, buffer_t,	sag_underlying_custom_t // complex types
		>;

	template<typename T>
	using is_gettable_t = typename std::enable_if<is_gettable<ampl::remove_const_t<ampl::remove_ref_t<T> > >::value, ampl::remove_ref_t<T> >::type;

	template<typename T> struct _type_name;
	template<> struct _type_name<int64_t> { static const char *name() { return "integer"; } };
	template<> struct _type_name<double> { static const char *name() { return "double"; } };
	template<> struct _type_name<bool> { static const char *name() { return "boolean"; } };
	template<> struct _type_name<const char*> { static const char *name() { return "string"; } };
	template<> struct _type_name<std::string> { static const char *name() { return "string"; } };
	template<> struct _type_name<map_t> { static const char *name() { return "map"; } };
	template<> struct _type_name<list_t> { static const char *name() { return "list"; } };
	template<> struct _type_name<decimal_t> { static const char *name() { return "decimal"; } };
	template<> struct _type_name<buffer_t> { static const char *name() { return "buffer"; } };
	template<> struct _type_name<sag_underlying_custom_t>  { static const char *name() { return "custom"; } };

	template<typename T>
	using type_name = _type_name<ampl::remove_const_t<ampl::remove_ref_t<T> > >;


	/**
	 * \internal
	 *
	 * The visitor implementation of get<T>
	 */
	template<typename T>
	struct GetVisitor
	{
		/** \internal. The result type */
		// If tag is undefined type then you are trying to use get with incorrect type,
		// get should only be used with one of the member type of data_t
		using result_type = is_gettable_t<ampl::remove_ref_t<T> > &;
		/** \internal. Throw on empty */
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		/** \internal. Return the value if it's the correct type */
		result_type operator()(ampl::remove_const_t<ampl::remove_ref_t<T> > &t) const { return t; }
		/** \internal. Return the value if it's the correct type */
		result_type operator()(const ampl::remove_const_t<ampl::remove_ref_t<T> > &t) const { return t; }
		/** \internal. Throw on other types */
		template<typename OTHER>
		result_type operator()(OTHER &o) const {
			fprintf(stderr, "GetVisitor<%s>.operator()(%s)\n", typeid(T).name(), typeid(T).name());
			throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<T>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }
	};

	template<>
	struct GetVisitor<const std::string &>
	{
		typedef std::string result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(const char *s) const { return s; }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<std::string>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }
	};

	template<>
	struct GetVisitor<const std::string>
	{
		typedef std::string result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(const char *s) const { return s; }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<std::string>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }
	};

	template<>
	struct GetVisitor<std::string >
	{
		typedef std::string result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(const char *s) const { return s; }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<std::string>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }
	};



	template<typename T>
	struct GetVisitor<custom_t<T> &>
	{
		typedef custom_t<T> &result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<sag_underlying_custom_t>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }

	};


	template<typename T>
	struct GetVisitor<const custom_t<T> >
	{
		typedef const custom_t<T> &result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		result_type operator()(const sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<sag_underlying_custom_t>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }

	};


	template<typename T>
	struct GetVisitor<custom_t<T> >
	{
		typedef custom_t<T> &result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<sag_underlying_custom_t>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }

	};

	template<typename T>
	struct GetVisitor<const custom_t<T> &>
	{
		typedef const custom_t<T> &result_type;
		result_type operator()() const { throw std::runtime_error("Get from empty"); }
		result_type operator()(sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		result_type operator()(const sag_underlying_custom_t &t) const { return reinterpret_cast<result_type>(t); }
		template<typename OTHER>
		result_type operator()(OTHER &o) const { throw std::runtime_error("Incorrect type in get (you asked for "+std::string(type_name<sag_underlying_custom_t>::name())+" but it's actually "+std::string(type_name<OTHER>::name())+")"); }

	};


	// only exists to deduce result_type, should not be called
	template<>
	struct GetVisitor<data_t&>
	{
		typedef data_t &result_type;
	};
	template<>
	struct GetVisitor<const data_t&>
	{
		typedef const data_t &result_type;
	};


} // get_details

// implementation of convert_to (below)
namespace convert_to_details {

	// stringstreams are very slow when used currently, thus we have these helper methods for converting numeric types to strings
	inline std::string integerToString(int64_t num)
	{
		char result[25];
		snprintf(result, sizeof(result), "%" PRId64, num);
		return std::string(result);
	}

	inline std::string doubleToString(double num)
	{
		char result[50];
		snprintf(result, sizeof(result), "%g", num);
		return std::string(result);
	}

	/** \internal
	 * Tag of types which can be parsed.
	 */
	template<class T>
	using is_convertable = ampl::is_one_of<T, int64_t, double, bool, std::string>;
	template<class T>
	using is_convertable_t = typename std::enable_if<is_convertable<T>::value, T>::type;
	template<class T>
	using is_not_convertable_t = typename std::enable_if<!is_convertable<T>::value, T>::type;
	
	/** \internal it's the correct type, just return it */
	template<typename T, typename OTHER> 
	inline ampl::is_same_t<T, OTHER> try_cast(const char *,const OTHER &t) { return t; }
	/** \internal it's the wrong type, throw */
	template<typename T, typename OTHER>
	inline ampl::is_not_same_t<T, OTHER> try_cast(const char *type_name, const OTHER &s) { throw std::runtime_error("Trying to parse from incorrect type (actually "+std::string(type_name)+")"); }
	template<typename T>
	inline T try_cast(const char *, const char *s);
	/** \internal int from a string */
	template<>
	inline int64_t try_cast<int64_t>(const char *, const char *s)
	{
		return atoll(s);
	}
	/** \internal double from a string */
	template<>
	inline double try_cast<double>(const char *, const char *s)
	{
		return atof(s);
	}
	/** \internal bool from a string */
	template<>
	inline bool try_cast<bool>(const char *, const char *s) 
	{
		std::string str(s);
		std::transform(str.begin(), str.end(), str.begin(), ::tolower);
		if (str == "true") return true;
		if (str == "false") return false;
		throw std::runtime_error("Invalid boolean: "+str);
	}
	/** \internal string from double */
	template<>
	inline std::string try_cast<std::string, double>(const char *, const double &t)
	{
		return doubleToString(t);
	}
	/** \internal string from int64_t */
	template<>
	inline std::string try_cast<std::string, int64_t>(const char *, const int64_t &t)
	{
		return integerToString(t);
	}
	/** \internal string from bool */
	template<>
	inline std::string try_cast<std::string, bool>(const char *, const bool &t)
	{
		return t ? "true" : "false";
	}
	/** \internal string from string */
	template<>
	inline std::string try_cast<std::string, char*>(const char *, char *const &t)
	{
		return t;
	}
	/** \internal string from string */
	template<>
	inline std::string try_cast<std::string>(const char *, const char *t)
	{
		return t;
	}

	/** \internal visitor implementation for convert_to */
	template<typename T>
	struct _get_or_convert
	{
		_get_or_convert(): type_name("") {}
		_get_or_convert(const char *type_name): type_name(type_name) {}
		typedef T result_type;
		is_convertable_t<T> operator()() const { throw std::runtime_error("Trying to parse from empty"); }
		template<typename OTHER>
		is_convertable_t<T> operator()(const OTHER &o) const
		{
			return try_cast<T>(type_name, o); 
		}
		const char *type_name;
	};

} // convert_to_details

template<typename T>
using is_convertable_t = convert_to_details::is_convertable_t<T>;
template<typename T>
using is_not_convertable_t = convert_to_details::is_not_convertable_t<T>;

#endif // _SAG_INTERNAL_CONVERT_TO_HPP_
